package de.landsh.opendata.catalogproxy;

import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.jena.irix.IRIException;
import org.apache.jena.irix.IRIs;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;

public class FilterInvalidRDF extends InputStream {


    private static final Logger log = LoggerFactory.getLogger(FilterInvalidRDF.class);

    private final BufferedReader reader;
    boolean initialized = false;
    private byte[] currentLine = null;
    private int index = 0;

    public FilterInvalidRDF(InputStream inputStream) {
        reader = new BufferedReader(new InputStreamReader(inputStream, StandardCharsets.UTF_8));
    }

    /**
     * Replace invalid information, such as IRIs with spaces and unescaped ampersand characters.
     */
    static String filterLine(String line) {
        if (line == null) return null;

        if (line.contains("&")) {
            String[] fragments = line.split("&");
            for (int i = 1; i < fragments.length; i++) {
                String fragment = fragments[i];
                if (!(fragment.startsWith("#") || fragment.startsWith("amp") || fragment.startsWith("apos")
                        || fragment.startsWith("quot") || fragment.startsWith("lt") || fragment.startsWith("gt"))) {
                    // invalid character entity reference
                    fragments[i] = "amp;" + fragment;
                }
            }
            line = StringUtils.join(fragments, "&");
        }

        line = fixRdfAbout(line, "about");
        return fixRdfAbout(line, "resource");
    }

    private static String fixRdfAbout(String line, String propertyName) {
        final String propertyWithDoubleQuote = "rdf:" + propertyName + "=\"";
        final String propertyWithSingleQuote = "rdf:" + propertyName + "='";

        String before = null;
        String iri = null;
        String after = null;
        if (line.contains(propertyWithDoubleQuote)) {
            before = StringUtils.substringBefore(line, propertyWithDoubleQuote) + propertyWithDoubleQuote;
            iri = StringUtils.substringBetween(line, propertyWithDoubleQuote, "\"");
            after = "\"" + StringUtils.substringAfter(StringUtils.substringAfter(line, propertyWithDoubleQuote), "\"");
        } else if (line.contains(propertyWithSingleQuote)) {
            before = StringUtils.substringBefore(line, propertyWithDoubleQuote) + propertyWithSingleQuote;
            iri = StringUtils.substringBetween(line, propertyWithDoubleQuote, "'");
            after = "'" + StringUtils.substringAfter(StringUtils.substringAfter(line, propertyWithDoubleQuote), "'");
        }

        if (iri == null) {
            return line;
        } else {

            return before + fixIRI(iri) + after;
        }
    }

    private static String fixIRI(String iri) {
        String result = iri.replaceAll(" ", "%20");

        try {
            IRIs.reference(result);
        } catch (IRIException e) {
            log.info("Found a really bad IRI: '{}'", iri);

            // A reliable way to turn any text (even a broken URI) into a URI is to use a hash value.

            result = "urn:" + DigestUtils.md5Hex(iri);
        }

        return result;
    }

    private void readNextLine() throws IOException {
        final String line = reader.readLine();
        if (line == null) {
            currentLine = null;
        } else {
            currentLine = filterLine(line).getBytes(StandardCharsets.UTF_8);
        }

        index = 0;
    }

    @Override
    public int read() throws IOException {
        if (!initialized) {
            initialized = true;
            readNextLine();

        }

        if (currentLine == null) {
            return -1;
        }

        if (index > currentLine.length) {
            readNextLine();
        }

        if (currentLine == null) {
            return -1;
        }
        // insert a newline character at the end of each line
        if (index == currentLine.length) {
            index++;
            return '\n';
        }

        byte result = currentLine[index];
        index++;
        return result;
    }
}
